/*
 * SerialStream
 * Copyright (c) 2004 by Andreas Schroeder <andreas@a-netz.de>
 *
 * This file is part of SerialStream.
 *
 * SerialStream is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * SerialStream is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SerialStream; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using System.IO;
using System.Runtime.InteropServices;

namespace System.IO.Serial
{
	/// <summary>
	/// This class implements a stream reading data from and writing data
	/// to a serial device. 
	/// </summary>
	public class SerialStream : System.IO.Stream
	{
		private FileStream iostream = null;
		private int port = 1;				
		private uint timeout = 1000;
		private int baudrate = 9600;
		private ParityEnum parity = ParityEnum.None;
		private StopBitEnum stopbits = StopBitEnum.One;
		private int databits = 8;
		private int buffersize = 512;
		private System.IntPtr hComm = (IntPtr) 0;

		public SerialStream(int _port, 
			int _baudrate, 
			ParityEnum _parity, 
			StopBitEnum _stopbits, 
			int _databits)
		{
			port = _port;
			baudrate = _baudrate;
			parity = _parity;
			stopbits = _stopbits;
			databits = _databits;

			Open();
		}

		private enum PurgeBuffers
		{
			TxAbort = 0x01,
			RXAbort = 0x02,
			TxClear = 0x04,
			RXClear = 0x08,
		}

		private enum Lines
		{
			SetRts = 3,
			ClearRts = 4,
			SetDtr = 5,
			ClearDtr = 6,
		}

		public System.UInt32 Timeout
		{
			get
			{
				return timeout;
			}
			set
			{
				timeout = (value == 0) ? 500 : value;
			}
		}

		public ParityEnum Parity
		{
			get
			{
				return parity;
			}
			set
			{
				parity = value;
			}
		}

		public StopBitEnum  StopBit
		{
			get
			{
				return stopbits;
			}
			set
			{
				stopbits = value;
			}
		}
	

		// Comunication Baudrate
		public System.Int32 BaudRate
		{
			get
			{
				return baudrate;
			}
			set
			{
				baudrate = value;
			}
		}

		// Comunication DataBits
		public System.Int32 DataBits
		{
			get
			{
				return databits;
			}
			set
			{
				databits = value;
			}
		}

		private void Open()
		{
			// Get Dcb block,Update with current data
			Win32Comm.DCB uDcb = new Win32Comm.DCB();

			System.Int32 iRc;

			if( !(port > 0) )
			{
				throw new ArgumentException("Port Number must be greater zero");
			}


			try
			{		
				if(iostream != null)
				{
					Close();
				}

				hComm = Win32Comm.CreateFile("COM" + port.ToString() + ":", 
					(uint)Win32Comm.AccessModeEnum.GenericRead + (uint)Win32Comm.AccessModeEnum.GenericWrite,
					0, //(uint) Win32Comm.FileShareModeEnum.Read, 
					0, 
					(uint) Win32Comm.CreateModeEnum.OpenExisting, 
					//(uint) Win32Comm.FileAttributeEnum.Normal, 
					0x80,
					0);

				// Clear all comunication errors
				int lpErrCode = 0;
				iRc = Win32Comm.ClearCommError(hComm, lpErrCode, 0);

				// Clears I/O buffers
				iRc = Win32Comm.PurgeComm(hComm, (int) (PurgeBuffers.RXClear | PurgeBuffers.TxClear) );

				// Gets COM Settings
				iRc = Win32Comm.GetCommState(hComm, ref uDcb);

				// Updates COM Settings
				char[] cParity = new char[]{'n', 'o', 'e', 'm'};
				string sDCBState = baudrate.ToString() 
					+ "," + cParity[(int)parity] 
					+ "," + databits.ToString() 
					+ "," + Convert.ToInt32(stopbits).ToString();

				iRc = Win32Comm.BuildCommDCB(sDCBState, ref uDcb);
				iRc = Win32Comm.SetCommState(hComm, ref uDcb);

				// Setup Buffers (Rx,Tx)
				iRc = Win32Comm.SetupComm(hComm, buffersize, buffersize);

				// Set Timeouts
				SetTimeout();

				// Creates a COM Port stream handle
				iostream = new FileStream((IntPtr) hComm, FileAccess.Read);
			}
			catch(Exception Ex)
			{
				throw new Exception("open port: " + Ex.Source + " - " + Ex.Message);
			}
		}

		public override void Close()
		{
			// Closes COM Port
			try
			{
				if(iostream != null)
				{
					iostream.Flush();
					iostream.Close();
					iostream = null;
				}
			}
			catch(Exception ex)
			{
				throw ex;
			}
			finally
			{
				Win32Comm.CloseHandle(hComm);
				iostream = null;
				hComm = (IntPtr) 0;
			}
		}
	
		public bool IsOpen
		{
			get
			{
				return (iostream != null);
			}
		}
		
		public override int Read(byte[] data, int offset, int count)
		{
			if(!IsOpen)
			{
				throw new Exception("Please initialize and open port before using this method");
			}

			try
			{
				return iostream.Read(data, offset, count);
			}
			catch(Exception ex)
			{
				throw new Exception("RX Error: " + ex.Message);
			}
		}

		public override void Write(byte[] data, int offset, int count)
		{
			if(!IsOpen)
			{
				throw new Exception("Please initialize and open port before using this method");
			}

			try
			{
				iostream.Write(data, offset, count);
			}
			catch(Exception ex)
			{
				throw new Exception("RX Error: " + ex.Message);
			}
		}

		public bool Rts
		{
			set
			{
				if(!IsOpen)
				{
					throw new Exception("Port is not opened.");
				}

				if(value)
				{
					Win32Comm.EscapeCommFunction(hComm, (int) Lines.SetRts);
				}
				else
				{
					Win32Comm.EscapeCommFunction(hComm, (int) Lines.ClearRts);
				}
			}
		}

		public bool Dtr
		{
			set
			{
				bool bRc;
				if(!IsOpen)
				{
					throw new Exception("Port is not opened.");
				}

				if(value)
				{
					bRc = Win32Comm.EscapeCommFunction(hComm, (int) Lines.SetDtr);
				}
				else
				{
					bRc = Win32Comm.EscapeCommFunction(hComm, (int) Lines.ClearDtr);
				}
			}
		}

		private void SetTimeout()
		{
			Win32Comm.COMMTIMEOUTS uCtm;
			int iRc;
			
			if(hComm==(IntPtr)0)
			{
				throw new Exception("Error while opening port and setting timeouts.");
			}

			// return with any received character or wait 
			// miTimeout ms for an character to receive
			// See msdn COMMTIMEOUTS doc.
			uCtm.ReadIntervalTimeout = 0xffffffff;
			uCtm.ReadTotalTimeoutMultiplier = 0xffffffff;
			uCtm.ReadTotalTimeoutConstant = timeout;
			// no timeout forwrite ops
			uCtm.WriteTotalTimeoutMultiplier = 0;
			uCtm.WriteTotalTimeoutConstant = 0;

			iRc = Win32Comm.SetCommTimeouts(hComm, ref uCtm);
		}

		public override bool CanRead
		{
			get
			{
				return true;
			}
		}

		public override bool CanWrite
		{
			get
			{
				return true;
			}
		}

		public override bool CanSeek
		{
			get
			{
				// serial streams do not support seeking
				return false;
			}
		}

		public override void Flush()
		{
			if(!IsOpen)
			{
				throw new Exception("Port is not opened.");
			}

			iostream.Flush();

			// Clears I/O buffers
			Win32Comm.FlushFileBuffers(hComm);
		}

		public override long Length
		{
			get
			{
				throw new NotSupportedException("Serial streams have no length.");
			}
		}

		public override long Position
		{
			get
			{
				throw new NotSupportedException("Serial streams do not support seeking");
			}
			set
			{
				throw new NotSupportedException("Serial streams do not support seeking");
			}
		}

		public override long Seek(long offset, System.IO.SeekOrigin origin)
		{
			throw new NotSupportedException("Serial streams do not support seeking");
		}

		public override void SetLength(long length)
		{
			throw new NotSupportedException("Serial streams have no length.");
		}

		public override int ReadByte()
		{
			if(!IsOpen)
			{
				throw new Exception("Port is not opened.");
			}

			return iostream.ReadByte();
		}

		public int Port
		{
			get
			{
				return port;
			}
		}


		// inner class to encapsulate win32 structs and calls 
		private class Win32Comm
		{
			// Win32 API
			[DllImport("kernel32.dll")] public extern static System.Int32 SetCommState(
				System.IntPtr hCommDev, 
				ref DCB lpDCB);
	
			[DllImport("kernel32.dll")] public extern static System.Int32 GetCommState(
				System.IntPtr hCommDev, 
				ref DCB lpDCB);

			[DllImport("kernel32.dll", CharSet=CharSet.Auto)] public extern static System.Int32 BuildCommDCB(
				string lpDef, 
				ref DCB lpDCB);

			[DllImport("kernel32.dll")] public extern static System.Int32 SetupComm(
				System.IntPtr hFile, 
				System.Int32 dwInQueue, 
				System.Int32 dwOutQueue);

			[DllImport("kernel32.dll")] public extern static System.Int32 SetCommTimeouts(
				System.IntPtr hFile, 
				ref COMMTIMEOUTS lpCommTimeouts);

			[DllImport("kernel32.dll")] public extern static System.Int32 GetCommTimeouts(
				System.IntPtr hFile, 
				ref COMMTIMEOUTS lpCommTimeouts);

			[DllImport("kernel32.dll")] public extern static System.Int32 ClearCommError(
				System.IntPtr hFile, 
				System.Int32 lpErrors, 
				System.Int32 l);

			[DllImport("kernel32.dll")] public extern static System.Int32 PurgeComm(
				System.IntPtr hFile, 
				System.Int32 dwFlags);

			[DllImport("kernel32.dll")] public extern static System.Boolean EscapeCommFunction(
				System.IntPtr hFile, 
				System.Int32 ifunc);

			[DllImport("kernel32.dll")] public extern static System.IntPtr CreateFile(
				[MarshalAs(UnmanagedType.LPStr)]
					string lpFileName,					// file name
				System.UInt32 dwDesiredAccess,			// access mode
				System.UInt32 dwShareMode,				// share mode
				System.UInt32 lpSecurityAttributes,		// SD
				System.UInt32 dwCreationDisposition,	// how to create
				System.UInt32 dwFlagsAndAttributes,		// file attributes
				System.UInt32 hTemplateFile);			// handle to template file

			[DllImport("kernel32.dll")] public extern static System.Int16 CloseHandle(
				System.IntPtr hObject);					// the handle to close

			[DllImport("kernel32.dll")] public extern static System.Int16 FlushFileBuffers(
				System.IntPtr hObject);					// the handle to close

			[DllImport("kernel32.dll")] public extern static System.Int16 ReadFile(
				System.IntPtr hFile,                // handle to file
				[MarshalAs(UnmanagedType.LPArray)] ref byte[] lpBuffer,             // data buffer
				System.Int32 nNumberOfBytesToRead,  // number of bytes to read
				ref System.Int32 lpNumberOfBytesRead, // number of bytes read
				System.IntPtr lpOverlapped    // overlapped buffer
				);

			public enum AccessModeEnum : uint
			{
				GenericRead = 0x80000000,
				GenericWrite = 0x40000000,
			}


			public enum FileShareModeEnum
			{
				Read = 0x00000001,
				Write = 0x00000002,
				Delete = 0x00000004,
			}

			public enum CreateModeEnum
			{
				CreateNew = 1,
				CreateAlways = 2,
				OpenExisting = 3,
			}

			public enum FileAttributeEnum
			{
				Normal = 0x80,
			}

			[StructLayout(LayoutKind.Sequential, Pack=1)] public struct DCB
			{
				private System.Int32 DCBlength;
				private System.Int32 BaudRate;
				private System.Int32 Bits1;
				private System.Int16 wReserved;
				private System.Int16 XonLim;
				private System.Int16 XoffLim;
				private System.Byte ByteSize;
				private System.Byte Parity;
				private System.Byte StopBits;
				private System.Byte XonChar;
				private System.Byte XoffChar;
				private System.Byte ErrorChar;
				private System.Byte EofChar;
				private System.Byte EvtChar;
				private System.Int16 wReserved2;
			}

			[StructLayout(LayoutKind.Sequential, Pack=1)] public struct COMMTIMEOUTS
			{
				public System.UInt32 ReadIntervalTimeout;
				public System.UInt32 ReadTotalTimeoutMultiplier;
				public System.UInt32 ReadTotalTimeoutConstant;
				public System.UInt32 WriteTotalTimeoutMultiplier;
				public System.UInt32 WriteTotalTimeoutConstant;
			}

			[StructLayout(LayoutKind.Sequential, Pack=1)] public struct COMMCONFIG
			{
				private System.Int32 dwSize;
				private System.Int16 wVersion;
				private System.Int16 wReserved;
				private DCB dcbx;
				private System.Int32 dwProviderSubType;
				private System.Int32 dwProviderOffset;
				private System.Int32 dwProviderSize;
				private System.Byte wcProviderData;
			}
		}
	}
}
