using System;
using System.Xml;

namespace SharpUtils.Config
{
	/// <summary>
	/// Diese Klasse dient zur Verwaltung von Einstellungen eines Programms. 
	/// 
	/// Es wird eine xml-Datei mit Einstellungen beim Erstellen eines Objekts
	/// dieser Klasse geladen. Wird auf eine Einstellung mit GetString, GetInt, ...
	/// zugegriffen, die noch nicht existiert, so wird ein entsprechender Eintrag in
	/// der xml-Datei angelegt und mit dem default-Wert besetzt.
	/// 
	/// ber die Methode Save werden diese nderungen in der xml-Datei gespeichert.
	/// </summary>
	public class ConfigManager
	{
		private XmlDocument configdoc;
		private string configname;
		private string dir;

		/// <summary>
		/// Konstruktor des ConfigManagers
		/// Es wird automatisch die xml-Datei mit dem Namen 
		/// [_configname].xml im Verzeichnis _dir geladen. 
		/// </summary>
		/// <param name="_dir">
		///   Verzeichnis, in dem nach der Configuration gesucht wird
		/// </param>
		/// <param name="_configname">Name der Konfiguration</param>
		public ConfigManager(string _configname)
		{
			configdoc = new XmlDocument();
			configname = _configname;
			dir = System.Windows.Forms.Application.StartupPath;

			Load();
		}

		private void Load()
		{
			try
			{
				configdoc.Load(dir + "\\" + configname + ".xml");
			}
			catch(Exception ex)
			{
				System.Windows.Forms.MessageBox.Show(
					"Cannot load configuration '" + configname + "': \n" 
					+ ex.Message + "\n"
					+ "Using default configuration.",
					"Configuration Manager",
					System.Windows.Forms.MessageBoxButtons.OK,
					System.Windows.Forms.MessageBoxIcon.Information);

				configdoc.AppendChild(
					configdoc.CreateProcessingInstruction("xml", "version=\"1.0\" encoding=\"utf-8\""));
				configdoc.AppendChild(
					configdoc.CreateElement("XmlSettings"));
			}
		}

		/// <summary>
		/// Gibt eine Einstellung in String-Form, zum Beispiel einen Hostnamen zurck
		/// </summary>
		/// <param name="_path">Pfad ("Name") der Einstellung</param>
		/// <param name="_default">Default-Wert</param>
		/// <returns></returns>
		public string GetString(string _path, string _default)
		{
			#if TRACE
				System.Diagnostics.Debug.WriteLine("Setting " + _path + " with default " + _default + " loaded.");
			#endif

			XmlElement element = GetPathElement(_path, true);

			string val = element.GetAttribute("Value");
			if(val != null && val != "")
			{
#if TRACE
				System.Diagnostics.Debug.WriteLine("Value " + val + " loaded.");
#endif
				return val;
			}
			else
			{
				XmlAttribute attribute = configdoc.CreateAttribute("Value");
				attribute.Value = _default;
				element.Attributes.Append(attribute);

#if TRACE
				System.Diagnostics.Debug.WriteLine("Default Value " + _default + " loaded.");
#endif

				return _default;
			}
		}

		/// <summary>
		/// Speichert einen String fr eine Einstellung
		/// </summary>
		/// <param name="_path">Einstellungs-Pfad</param>
		/// <param name="_value">Einstellungs-Wert</param>
		public void SetString(string _path, string _value)
		{
			XmlElement element = GetPathElement(_path, true);

			XmlAttribute attribute = element.GetAttributeNode("Value");
			if(attribute == null)
			{
				attribute = configdoc.CreateAttribute("Value");
				element.Attributes.Append(attribute);
			}

			attribute.Value = _value;
		}

		private XmlElement GetPathElement(string _path, bool _create)
		{
			XmlElement element = configdoc.DocumentElement;

			if(element != null && element.Name!= "XmlSettings")
			{
				throw new Exception("The supplied file is no xml setting file.");
			}

			string[] path = _path.Split(new char[]{'/'});

			XmlElement nextelement;
			foreach(string part in path)
			{
				if(part=="")
				{
					continue;
				}

				nextelement = (XmlElement)element.SelectSingleNode(part);

				if(nextelement == null)
				{
					if(_create)
					{
						// create it
						nextelement = configdoc.CreateElement(part);
						element.AppendChild(nextelement);
					}
					else
					{
						return null;
					}
				}

				element = nextelement;
			}

			return element;
		}

		/// <summary>
		/// Gibt zurck, ob eine Einstellung schon gespeichert wurde. 
		/// </summary>
		/// <param name="_path"></param>
		/// <returns></returns>
		public bool IsSet(string _path)
		{
			return ( GetPathElement(_path, false) != null );
		}
	
		/// <summary>
		/// Speichert die xml-Datei
		/// </summary>
		public void Save()
		{
			configdoc.Save(dir + "\\" + configname + ".xml");
		}

		/// <summary>
		/// Gibt eine Integer-Einstellung zurck
		/// </summary>
		/// <param name="_path"></param>
		/// <param name="_default"></param>
		/// <returns></returns>
		public int GetInt(string _path, int _default)
		{
			return Convert.ToInt32(
				GetString(_path, _default.ToString()));
		}

		/// <summary>
		/// Gibt eine bool-Einstellung zurck
		/// </summary>
		/// <param name="_path"></param>
		/// <param name="_default"></param>
		/// <returns></returns>
		public bool GetBool(string _path, bool _default)
		{
			return Convert.ToBoolean(
				GetString(_path, _default.ToString()));
		}

		public string Filename
		{
			get
			{
				return dir + configname + ".xml";
			}
		}
	}
}
