using System;

namespace SharpUtils
{
	/// <summary>
	/// This is a simple linked List
	/// </summary>
	public class LinkedList : System.Collections.ICollection
	{
		private ListItem m_head = null;
		private ListItem m_tail = null;
		private int m_count = 0;

		public void Clear() 
		{
			m_head = null;
			m_tail = null;
			m_count = 0;
		}


		/// <summary>
		/// The number of elements in the list
		/// </summary>
		public int Count
		{
			get
			{
				return m_count;
			}
		}

		/// <summary>
		/// Copies the content of the list to an array
		/// </summary>
		/// <param name="array"></param>
		/// <param name="index"></param>
		public void CopyTo(System.Array array, int index)
		{
			int pos = index;
			foreach(object o in this)
			{
				array.SetValue(o, pos++);
			}
		}

		/// <summary>
		/// Returns whether the list is able to handle access by multiple threads (is not)
		/// </summary>
		public bool IsSynchronized
		{
			get
			{
				return false;
			}
		}

		/// <summary>
		/// A synchronization object for concurrent access.
		/// </summary>
		public object SyncRoot
		{
			get
			{
				return null;
			}
		}

		/// <summary>
		/// Adds an item to the end of the list
		/// </summary>
		/// <param name="Item"></param>
		public void Add(object Item)
		{
			if(m_tail == null)
			{
				m_head = new ListItem( Item );
				m_tail = m_head;
			} 
			else
			{
				m_tail.next = new ListItem( Item );
				m_tail.next.prev = m_tail;
				m_tail = m_tail.next;
			}

			m_count++;
		}

		/// <summary>
		/// Adds an element to the end of the list (same as Add)
		/// </summary>
		/// <param name="Item"></param>
		public void Append(object Item)
		{
			Add( Item );
		}

		/// <summary>
		/// Adds an element to the beginning of the list
		/// </summary>
		/// <param name="Item"></param>
		public void Prepend(object Item)
		{
			if(m_head == null)
			{
				Add(Item);
			}
			else
			{
				ListItem item = new ListItem( Item );
				item.next = m_head;
				m_head.prev = item;
				m_head = item;
				m_count++;
			}
		}

		#region Implementation of IEnumerable
		/// <summary>
		/// Returns an enumerator of this list
		/// </summary>
		/// <returns></returns>
		public System.Collections.IEnumerator GetEnumerator()
		{
			return new ListEnumerator(m_head);
		}
		#endregion

		/// <summary>
		/// This is the container for the item data
		/// </summary>
		private class ListItem
		{
			public ListItem next = null;
			public ListItem prev = null;
			public object data;

			public ListItem(object _data)
			{
				data = _data;
			}
		}

		/// <summary>
		/// This is the enumerator returned by GetEnumerator
		/// </summary>
		private class ListEnumerator : System.Collections.IEnumerator
		{
			ListItem m_current;
			ListItem m_head;
		
			public ListEnumerator(ListItem _head)
			{
				m_head = _head;
				Reset();
			}

			#region Implementation of IEnumerator
			public void Reset()
			{
				m_current = null;
			}

			public bool MoveNext()
			{
				if(m_head == null && m_current == null)
				{
					return false;
				}
				if(m_current == null)
				{
					m_current = m_head;
					return true;
				}
				else if(m_current.next != null)
				{
					m_current = m_current.next;
					return true;
				}
				else
				{
					m_current = m_head;
				}
				return false;
			}

			public object Current
			{
				get
				{
					return m_current.data;
				}
			}
			#endregion
		}

		/// <summary>
		/// Removes the first element of the list
		/// </summary>
		/// <returns></returns>
		public object RemoveFirst()
		{
			if(m_head == null)
			{
				throw new Exception("Tried to removed element from empty list!");
			}
			else
			{
				object data = m_head.data;

				if(m_head.next == null)
				{
					m_head = null;
					m_tail = null;
				}
				else
				{
					m_head = m_head.next;
					m_head.prev = null;
				}

				m_count--;

				return data;
			}
		}

		/// <summary>
		/// Removes the last element of the list
		/// </summary>
		/// <returns></returns>
		public object RemoveLast()
		{
			if(m_tail == null)
			{
				throw new Exception("Tried to remove element from empty list!");
			}
			else
			{
				object data = m_tail.data;
				if(m_tail.prev == null)
				{
					m_head = null;
					m_tail = null;
				}
				else
				{
					m_tail = m_tail.prev;
					m_tail.next = null;
				}

				m_count--;

				return data;
			}
		}	

		/// <summary>
		/// first element of the list
		/// </summary>
		public object Head
		{
			get
			{
				if(m_head != null)
				{
					return m_head.data;
				}
				else
				{
					return null;
				}
			}
		}

		/// <summary>
		/// Last element of the list
		/// </summary>
		public object Tail
		{
			get
			{
				if(m_tail!=null)
				{
					return m_tail.data;
				}
				else
				{
					return null;
				}
			}
		}

		public object Remove(Object data)
		{
			ListItem item = m_head;
			if(m_head == null)
				return null;
			
			while(item.data != null)
			{
				if(item.data == data)
				{
					if(item == m_head)
					{
						RemoveFirst();
						return item.data;
					} 
					else if(item == m_tail)
					{
						RemoveLast();
						return item.data;
					} 
					else
					{
						//Object d = item.data;
						item.prev.next = item.next;
						item.next.prev = item.prev;
						return item.data;
					}
				}
				item = item.next;
			}
			return null;
		}
	}
}
