/*
 * VC820Tool
 * Copyright (c) 2004 by Andreas Schroeder <andreas@a-netz.de>
 *
 * This file is part of VC820Tool.
 *
 * VC820Tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * VC820Tool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VC820Tool; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

namespace VC820Tool
{
	/// <summary>
	/// This class decodes and stores the decoded data of one 
	/// data packet received from the instrument
	/// </summary>
	public class Datagram
	{
		private byte[] input;

		public bool Auto;
		public bool AC;
		public bool TrueRms;
		public bool RS232;

		public bool Diode;
		public bool Hold;
		public bool Delta;
		public bool Beep;
		
		public bool Overload;

		public PrefixEnum Prefix;
		public UnitsEnum Unit;

		public decimal Value;

		public DateTime Timestamp;
		
		public Datagram(byte[] _input)
		{
			input = _input;
			DecodeInput();
			Timestamp = DateTime.Now;
		}

		public Datagram(UnitsEnum _unit, PrefixEnum _prefix, decimal _value)
		{
			Unit = _unit;
			Prefix = _prefix;
			Value = _value;

			Timestamp = DateTime.Now;
		}

		protected void DecodeInput()
		{	
			DecodeBits();
			DecodeUnit();
			DecodeOverload();
			DecodeValue();
			DecodePrefix();
		}

		protected void DecodeValue()
		{
			if(Overload)
			{
				Value = 0;
			}
			else
			{
				Value = (decimal) (1000 * DecodeNumber((byte) (input[1] << 4 | (input[2] & 0x0f)))
					+ 100 * DecodeNumber((byte)(input[3] << 4 | (input[4] & 0x0f)))
					+ 10 * DecodeNumber((byte) (input[5]<<4 | (input[6] & 0x0f)))
					+ DecodeNumber((byte) (input[7]<<4 | (input[8] & 0x0f))))
					* GetFactor();
			}
		}

		protected byte DecodeNumber(byte number)
		{
			number &= 0x7f;

			switch(number)
			{
				case 0x7d: return 0; 
				case 0x05: return 1; 
				case 0x5b: return 2; 
				case 0x1f: return 3; 
				case 0x27: return 4; 
				case 0x3e: return 5; 
				case 0x7e: return 6; 
				case 0x15: return 7; 
				case 0x7f: return 8; 
				case 0x3f: return 9; 
				default:
					Console.WriteLine("Unknown number " + number.ToString("x"));
					return 0;
			}
		}

		protected decimal GetFactor()
		{
			decimal factor = 1;
			if((input[1] & 0x08) == 0x08)
				factor *= -1M;
			if((input[3] & 0x08) == 0x08)
				factor *= 0.001M;
			if((input[5] & 0x08) == 0x08)
				factor *= 0.01M;
			if((input[7] & 0x08) == 0x08)
				factor *= 0.1M;

			return factor;
		}
		protected void DecodeBits()
		{
			RS232 = GetBit(0, 0); // ???
			Auto = GetBit(0, 1);
			TrueRms = GetBit(0, 2); // ???
			AC = GetBit(0, 3);
			
			Diode = GetBit(9, 0);
			Beep = GetBit(10, 0);
			Hold = GetBit(11, 0);
			Delta = GetBit(11, 1);
		}

		protected bool GetBit(byte _byte, byte _bit)
		{
			byte bitmask = (byte) (1 << _bit);
			return( (input[_byte] & bitmask) == bitmask);
		}

		protected void DecodeUnit()
		{
			if( GetBit(11, 3) )
				Unit = UnitsEnum.Farad;
			else if( GetBit(11, 2) )
				Unit = UnitsEnum.Ohm;
			else if( GetBit(12, 3) )
				Unit = UnitsEnum.Ampere;
			else if( GetBit(12, 1) )
				Unit = UnitsEnum.Hertz;
			else if( GetBit(12, 2) )
				Unit = UnitsEnum.Volt;
			else if( GetBit(10, 2) )
				Unit = UnitsEnum.Prozent;
			else 
				Console.WriteLine("Unkown Unit");
		}

		protected void DecodePrefix()
		{
			if(GetBit(9, 2))
				Prefix = PrefixEnum.Nano;
			else if(GetBit(9, 3))
				Prefix = PrefixEnum.Micro;
			else if(GetBit(10, 3))
				Prefix = PrefixEnum.Milli;
			else if(GetBit(9, 1))
				Prefix = PrefixEnum.Kilo;
			else if(GetBit(10, 1))
				Prefix = PrefixEnum.Mega;
			else 
				Prefix = PrefixEnum.None;
		}

		protected void DecodeOverload()
		{
			Overload = ((input[3] & 0x07) ==0x07) 
				&& ((input[4] & 0x0f) ==0x0d) 
				&& ((input[5] & 0x07) ==0x06) 
				&& ((input[6] & 0x0f) ==0x08);
		}

		public override string ToString()
		{
			string output;
			output = string.Format("Value: {0}", Value);

			return output;
		}

		public static string GetUnitName(UnitsEnum unit)
		{
			switch(unit)
			{
				case UnitsEnum.Ampere:
					return "Ampere";
				case UnitsEnum.Farad:
					return "Farad";
				case UnitsEnum.GradCelsius:
					return "Grad Celsius";
				case UnitsEnum.Hertz:
					return "Hertz";
				case UnitsEnum.Ohm:
					return "Ohm";
				case UnitsEnum.Prozent:
					return "Prozent";
				case UnitsEnum.Volt:
					return "Volt";
				default:
					return "Unknown";
			
			}
		}

		public string PrefixName
		{
			get
			{
				switch(Prefix)
				{
					case PrefixEnum.Nano:
						return "Nano";
					case PrefixEnum.Micro:
						return "Mikro";
					case PrefixEnum.Milli:
						return "Milli";
					case PrefixEnum.None:
						return "";
					case PrefixEnum.Kilo:
						return "Kilo";
					case PrefixEnum.Mega:
						return "Mega";
					default:
						return "Unknown";
				}
			}
		}

		public decimal GetPrefixFactor()
		{
			return (decimal) Math.Pow(10, (double)Prefix);
		}

		public decimal RealValue
		{
			get
			{
				return Value * GetPrefixFactor();	
			}
		}

		public byte[] InternalData
		{
			get
			{
				return input;
			}
		}
	}
}
