/*
 * VC820Tool
 * Copyright (c) 2004 by Andreas Schroeder <andreas@a-netz.de>
 *
 * This file is part of VC820Tool.
 *
 * VC820Tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * VC820Tool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VC820Tool; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using SharpUtils;

namespace VC820Tool.Processors
{
	public class Statistics
	{
		public enum TendencyEnum
		{
			Rising,
			Falling,
			Equal,
		}

		private Datagram lastdatagram = null;
		private TendencyEnum tendency;
		private MaxLengthQueue buffer;
		private double arith_avg;
		private decimal minimum;
		private decimal maximum;
		private double stddev;

		public Statistics(int initalbufferlen)
		{
			buffer = new MaxLengthQueue(initalbufferlen);
		}

		public void OnNewDatagram(Datagram dgram)
		{
			DoTendency(dgram);
			DoBuffer(dgram);
			DoMinMax(dgram);
			DoStandardDeviation(dgram);
			lastdatagram = dgram;
		}

		private void DoTendency(Datagram datagram)
		{
			if(lastdatagram!=null)
			{
				if(lastdatagram.RealValue > datagram.RealValue)
				{
					tendency = TendencyEnum.Falling;
				}
				else if(lastdatagram.Value < datagram.Value)
				{
					tendency = TendencyEnum.Rising;
				}
				else if(lastdatagram.Value == datagram.Value)
				{
					tendency = TendencyEnum.Equal;
				}
			}
		}

		private void DoBuffer(Datagram datagram)
		{
			arith_avg = 0;

			if((lastdatagram != null) && (datagram.Unit != lastdatagram.Unit))
			{
				buffer.List.Clear();
			}

			buffer.Enqueue(datagram);
			
			foreach(Datagram dgram in buffer.List)
			{
				arith_avg += (double)dgram.RealValue;
			}

			arith_avg /= buffer.Count;
		}

		private void DoMinMax(Datagram datagram)
		{
			if(lastdatagram==null)
			{
				minimum = datagram.RealValue;
				maximum = datagram.RealValue;
			}

			if(minimum > datagram.RealValue)
			{
				minimum = datagram.RealValue;
			}

			if(maximum < datagram.RealValue)
			{
				maximum = datagram.RealValue;
			}
		}

		private void DoStandardDeviation(Datagram _dgram)
		{
			double sumsq = 0;

			if(buffer.Count == 0)
			{
				stddev = 0;
				return;
			}

			foreach(Datagram dgram in buffer.List)
			{
				sumsq += Math.Pow((((double)dgram.RealValue) - arith_avg), 2);
			}

			stddev = Math.Sqrt( (1 / ((double)buffer.Count)) * sumsq);
		}

		public void ResetMinMax()
		{
			if(lastdatagram != null)
			{
				maximum = lastdatagram.RealValue;
				minimum = lastdatagram.RealValue;
			}
		}

		public TendencyEnum Tendency
		{
			get
			{
				return tendency;
			}
		}

		public double ArithmethicAverage
		{
			get
			{
				return arith_avg;
			}
		}

		public int BufferLength
		{
			get
			{
				return buffer.Length;
			}
			set
			{
				buffer.Length = value;
			}
		}

		public int BufferCount
		{
			get
			{
				return buffer.Count;
			}
		}

		public decimal Minimum
		{
			get
			{
				return minimum;
			}
		}

		public decimal Maximum
		{
			get
			{
				return maximum;
			}
		}

		public double StandardDeviation
		{
			get
			{
				return stddev;
			}
		}
	}
}

