/*
 * VC820Tool
 * Copyright (c) 2004 by Andreas Schroeder <andreas@a-netz.de>
 *
 * This file is part of VC820Tool.
 *
 * VC820Tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * VC820Tool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VC820Tool; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using System.Threading;
using System.IO.Serial;

namespace VC820Tool.Sources
{
	/// <summary>
	/// This class reads the serial input received by the instrument.
	/// If a complete datagram (14 bytes) has been received, it constructs 
	/// a Datagram object an raises the NewDatagram event.
	/// </summary>
	public class SerialSource : AbstractDatagramSource, IDatagramSource
	{
		private byte[] inputbuffer = new byte[14];
		private int inputpos = 1;
		private SerialStream serial;
		private Thread readerthread = null;
		private SerialSourceCtrl ctrl;
		private int commport;

		public SerialSource()
		{
			ctrl = new SerialSourceCtrl(this);
			commport = Settings.CM.GetInt("Source/" + this.InternalName + "/Port", 1);
		}

		protected override void OnStart()
		{
			try
			{
				serial = new SerialStream(
					commport, 
					2400, 
					ParityEnum.None,
					StopBitEnum.One,
					8);

				serial.Dtr = true;
				serial.Rts = false;

				readerthread = new System.Threading.Thread(new System.Threading.ThreadStart(this.DoReading));
				readerthread.Name = "Serial Reader Thread";
				readerthread.Start();
				
			}
			catch(Exception ex)
			{
				OnStop();
				enabled = false;
				throw new Exception("Cannot open communications port.",	ex);
			}
		}

		protected override void OnStop()
		{
			try
			{	
				if(readerthread!=null)
				{
					readerthread.Abort();
				}
				if(serial.IsOpen)
				{
					serial.Dtr = false;
					serial.Rts = false;
					serial.Close();
				}
			}
			catch(Exception ex)
			{
				System.Diagnostics.Debug.WriteLine("Closing of Comm Port failed: " + ex.Message);
			}
		}

		private void DoReading()
		{
			int data = 0;

			while(enabled)
			{
				data = serial.ReadByte();
				if(data >= 0)
				{
					ProcessInput( (byte)data );
				}
			}
		}

		private void ProcessInput(byte input)
		{
			byte charpos = 0;

			charpos = (byte)((input & 0xf0) >> 4);

			if( charpos == inputpos)
			{
				inputbuffer[inputpos - 1] = input;
				inputpos++;
			}
			else
			{
				inputpos = 1;
			}

			if(inputpos == 15)
			{
				// datagram complete
				inputpos = 1;
				OnNewDatagram(new Datagram(inputbuffer));
			}
		}

		protected virtual void OnNewDatagram(Datagram dgram) 
		{
			if( vd != null )
			{
				vd.OnNewDatagram( this, dgram );
			}
		}

		public int Port
		{
			get
			{
				return serial.Port;
			}
			set
			{
				OnStop();
				
				commport = value;

				OnStart();
			}
		}

		#region Implementation of IDatagramSource
		public override string UIName
		{
			get
			{
				return "Serial VC820";
			}
		}

		public override string InternalName
		{
			get
			{
				return "SerialVC820";
			}
		}

		public override ControllingGui UserInterface
		{
			get
			{
				return ctrl;
			}
		}
		#endregion
	}
}
