/*
 * VC820Tool
 * Copyright (c) 2004 by Andreas Schroeder <andreas@a-netz.de>
 *
 * This file is part of VC820Tool.
 *
 * VC820Tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * VC820Tool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VC820Tool; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;
using SharpUtils;
using System.Threading;

namespace VC820Tool
{
	/// <summary>
	/// This class gets each datagram received from the instrument and 
	/// informs all registered DatagramProcessors about this
	/// datagram each time the current DispatcherTrigger calls Trigger(). 
	/// </summary>
	public class ValueDispatcher
	{
		private LinkedList processors;
		private Datagram lastdgram;
		private Thread procthread;
		private bool processing = false;
		private object proclock = new object();

		public event EventHandler DatagramLost;

		public ValueDispatcher()
		{
			processors = new LinkedList();

			procthread = new Thread(new ThreadStart(InternalTrigger));
			procthread.IsBackground = true;
			procthread.Name = "Value Dispatcher";
			procthread.Start();
		}

		public void OnNewDatagram(Sources.IDatagramSource source, Datagram dgram)
		{
			lastdgram = dgram;
			FireNewDatagram(source, dgram);
		}

		public void AddDatagramProcessor(IDatagramProcessor dgproc)
		{
			processors.Add(dgproc);
		}

		public void RemoveDatagramProcessor(IDatagramProcessor dgproc)
		{
			processors.Remove(dgproc);
		}

		public void Trigger()
		{
			if(lastdgram == null)
				return;

			lock(proclock)
			{
				if(processing) 
				{
					System.Diagnostics.Debug.WriteLine("Warning: Processing call overflow. Missing datagrams!");
					if(DatagramLost != null)
						DatagramLost(this, new EventArgs());
					return;
				}
				Monitor.Pulse(proclock);
			}
		}

		private void InternalTrigger()
		{
			while(true) 
			{
				lock(proclock)
				{
					Monitor.Wait(proclock);
					processing = true;
				}

				DateTime t1 = DateTime.Now;

				foreach(IDatagramProcessor proc in processors)
				{
					proc.OnDatagram(lastdgram);
				}

				TimeSpan span = DateTime.Now - t1;

				System.Diagnostics.Debug.WriteLine("Processors needed " + span.TotalMilliseconds + "ms.");

				lock(proclock)
				{
					processing = false;
				}
			}
		}

		private void FireNewDatagram(Sources.IDatagramSource source, Datagram dgram)
		{
			if(NewDatagram!=null)
			{
				NewDatagram(source, dgram);
			}
		}

		public event NewDatagramHandler NewDatagram;
	}

	public delegate void NewDatagramHandler(Sources.IDatagramSource source, Datagram dgram);
}
